/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of PeerfactSim.KOM.
 * 
 * PeerfactSim.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tud.kom.p2psim.impl.wsn;

import de.tud.kom.p2psim.api.energy.ComponentType;
import de.tud.kom.p2psim.api.energy.EnergyComponent;
import de.tud.kom.p2psim.api.energy.EnergyModel;
import de.tud.kom.p2psim.impl.common.Deactivatable;
import de.tud.kom.p2psim.impl.energy.components.OneStateEnergyComponent;
import de.tud.kom.p2psim.impl.energy.components.StateEnergyCommunicationComponent;
import de.tudarmstadt.maki.simonstrator.api.Event;
import de.tudarmstadt.maki.simonstrator.api.EventHandler;
import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Monitor;
import de.tudarmstadt.maki.simonstrator.api.Monitor.Level;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.HostComponent;
import de.tudarmstadt.maki.simonstrator.api.component.HostComponentFactory;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.SensorComponent;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.WSNSensorComponent;
import de.tudarmstadt.maki.simonstrator.api.util.XMLConfigurableConstructor;
import de.tudarmstadt.maki.simonstrator.api.Randoms;
import de.tudarmstadt.maki.simonstrator.api.Time;

/**
 * This class represents a generic sensor component
 * @author Lukas Wehrstein
 * @version 1.0, 31.01.2022
 */
public class GenericWSNSensor implements WSNSensorComponent, EnergyAwareHostComponent {

	private SensorType sensorType;
	private EnergyComponent energyComponent;
	private Host host;
	
	private boolean isActive;
	private String name;
	private double data;
	
	/**
	 * Creates a generic Sensor component
	 * @param host
	 * @param sensorType
	 * @param energyComponent for energy simulation
	 */
	public GenericWSNSensor(Host host, SensorType sensorType, EnergyComponent energyComponent) {
		this.sensorType = sensorType;
		this.energyComponent = energyComponent;
		this.host = host;
		this.isActive = false;
		this.name = sensorType.toString();
		this.data = 0;
		energyComponent.turnOff();
	}
	

	@Override
	public void initialize() {
		
		if(!isActive) {
			// sensor is turned off
			
		}
		try {
			EnergyModel energyModel = host.getComponent(EnergyModel.class);
			energyModel.registerComponent(this.energyComponent);
			
			if(energyComponent instanceof OneStateEnergyComponent) {
				OneStateEnergyComponent osec = (OneStateEnergyComponent) energyComponent;
				Monitor.log(getClass(), Level.DEBUG, "%s Sensor was registered with %s uJ", name, osec.getState().getEnergyConsumption());
			}
			
		} catch (ComponentNotAvailableException e) {
			Monitor.log(getClass(), Level.WARN, "Host %s: This host does not have a valid energy model", host.getId());
		}

	}

	@Override
	public void shutdown() {
		setActive(false);

	}

	@Override
	public Host getHost() {
		return this.host;
	}
	
	@Override
	public EnergyComponent getEnergyComponent() {
		return this.energyComponent;
	}
	
	/**
	 * EnergyComponent that represents this specific sensor component
	 * @param energyComponent
	 */
	@Override
	public void setEnergyComponent(EnergyComponent energyComponent) {
		this.energyComponent = energyComponent;
	}
	
	public SensorType getSensorType() {
		return this.sensorType;
	}
	
	
	@Override
	public boolean isActive() {
		return this.isActive;
	}


	@Override
	public void setActive(boolean b) {
		if(isActive != b) {
			// change state
			this.isActive = b;
			if(b) {
				getEnergyComponent().turnOn();
			}else {
				getEnergyComponent().turnOff();
			}
		}
	}
	

	@Override
	public boolean testSensor() {
		return true;
	}
	
	@Override
	public double getSensorData() {
//		setActive(true);
//		getEnergyComponent().turnOn();
//		generateRandomData();
//		
//		// TODO LW: Needs to be adjusted with propper timing solution...
//		//wait 1 second until data is there
//		Event.scheduleWithDelay(Time.SECOND, new EventHandler() {
//
//			@Override
//			public void eventOccurred(Object content, int type) {
//				getEnergyComponent().turnOff();
//				setActive(false);
//			}
//			
//		}, this, 0);
//		
//		return data;
		if(!isActive()) {
			setActive(true);
		}
		generateRandomData();
		return this.data;
	}
	
	/**
	 * Generates random sensor data
	 */
	public void generateRandomData() {
		this.data = Randoms.getRandom(getClass()).nextDouble();
		//data = Math.random()*2 - 1;
		Monitor.log(getClass(), Level.DEBUG, "Generated Random Double: ", this.data);
	}
	
	@Override
	public int getSensorDataSize() {
		return Double.BYTES;
	}
	
	@Override
	public String getName() {
		return name;
	}

	
	public void setName(String name) {
		this.name = name;
	}

	public static class Factory implements HostComponentFactory{

		SensorType sensorType;
		EnergyComponent energyComponent;
		double energyConsumptionUwatt;
		
		
		/**
		 * Type of the sensor
		 * @param sensorType
		 */
//		public void setSensorType(SensorType sensorType) {
//			System.out.println("set ST: " + sensorType);
//			this.sensorType = sensorType;
//		}
		
		public void setSensorType(String sensorType) {
			try {
				this.sensorType = SensorType.valueOf(sensorType.toUpperCase());
			}catch (Exception e) {
				Monitor.log(getClass(), Level.ERROR, "%s is not a valid SensorType. Using GENERIC instead", sensorType);
				this.sensorType = SensorType.GENERIC;
			}
			
			// System.out.println("set ST_string: " + sensorType);
			
		}
		
		
		/**
		 * Defines the energy consumption of this sensor
		 * @param energyConsumptionUwatt
		 */
		public void setEnergyConsumptionUwatt(double energyConsumptionUwatt) {
			Monitor.log(getClass(), Level.DEBUG, "EnergyConsumption of this %s sensor was set to %s uW", sensorType, this.energyConsumptionUwatt);
			this.energyConsumptionUwatt = energyConsumptionUwatt;
		}
		
		@Override
		public HostComponent createComponent(Host host) {
			this.energyComponent = new OneStateEnergyComponent(energyConsumptionUwatt, ComponentType.SENSOR, false);
			return new GenericWSNSensor(host, sensorType, energyComponent);
		}
		
	}

}
